"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const vscode_jsonrpc_1 = require("vscode-jsonrpc");
exports.CancellationToken = vscode_jsonrpc_1.CancellationToken;
exports.CancellationTokenSource = vscode_jsonrpc_1.CancellationTokenSource;
/**
 * Provides a token that is cancelled as soon as ALL added tokens are cancelled.
 * Useful for memoizing a function, where multiple consumers wait for the result of the same operation,
 * which should only be cancelled if all consumers requested cancellation.
 */
class CancellationTokenLink {
    constructor() {
        /**
         * Amount of total consumers
         */
        this.tokens = 0;
        /**
         * Amount of consumers that have requested cancellation
         */
        this.cancelled = 0;
        /**
         * Internal token source that is cancelled when all consumers request cancellation
         */
        this.source = new vscode_jsonrpc_1.CancellationTokenSource();
    }
    /**
     * A linked CancellationToken that is cancelled as soon as all consumers request cancellation
     */
    get token() {
        return this.source.token;
    }
    /**
     * Add another CancellationToken that needs to be cancelled to trigger cancellation
     *
     * @returns Disposable that allows to remove the token again
     */
    add(token) {
        this.tokens++;
        token.onCancellationRequested(() => {
            this.cancelled++;
            // If all consumers requested cancellation, cancel
            if (this.cancelled === this.tokens) {
                this.source.cancel();
            }
        });
    }
}
/**
 * Memoizes the result of a promise-returning function by the first argument with support for cancellation.
 * If the last argument is a CancellationToken, the operation is only cancelled if all calls have requested cancellation.
 * Rejected (or cancelled) promises are automatically removed from the cache.
 * If the operation has already finished, it will not be cancelled.
 *
 * @param func Function to memoize
 * @param cache A custom Map to use for setting and getting cache items
 *
 * @template F The function to be memoized
 * @template T The return type of the function, must be Promise
 * @template K The cache key (first argument to the function)
 */
function cancellableMemoize(func, cache = new Map()) {
    // Track tokens consumers provide
    const tokenLinks = new Map();
    const memoized = function (...args) {
        const key = args[0];
        // Get or create CancellationTokenLink for the given first parameter
        let tokenLink = tokenLinks.get(key);
        if (!tokenLink) {
            tokenLink = new CancellationTokenLink();
            tokenLinks.set(key, tokenLink);
        }
        // Take last argument as CancellationToken from arguments if provided or use a token that is never cancelled
        const token = vscode_jsonrpc_1.CancellationToken.is(args[args.length - 1]) ? args.pop() : vscode_jsonrpc_1.CancellationToken.None;
        // Add it to the list of tokens that need to be cancelled for final cancellation
        tokenLink.add(token);
        let result;
        // Check if function has been called with this argument already
        if (cache.has(key)) {
            // Return previous result
            result = cache.get(key);
        }
        else {
            // Call function
            // Pass the linked cancel token
            args.push(tokenLink.token);
            result = func.apply(this, args).catch(err => {
                // Don't cache rejected promises
                cache.delete(key);
                throw err;
            });
            // Save result
            cache.set(key, result);
        }
        return result;
    };
    return memoized;
}
exports.cancellableMemoize = cancellableMemoize;
/**
 * Thrown when an operation was cancelled
 */
class CancelledError extends Error {
    constructor(message = 'Operation cancelled') {
        super(message);
    }
}
exports.CancelledError = CancelledError;
/**
 * Returns true if the passed error was caused by a cancellation
 */
function isCancelledError(error) {
    return typeof error === 'object' && error !== null && (error instanceof CancelledError
        || error instanceof ts.OperationCanceledException
        || (typeof error.name === 'string' && error.name.includes('Cancel')));
}
exports.isCancelledError = isCancelledError;
/**
 * Throws a CancelledError if cancellation was requested for the passed CancellationToken
 */
function throwIfRequested(token) {
    if (token.isCancellationRequested) {
        throw new CancelledError();
    }
}
exports.throwIfRequested = throwIfRequested;
/**
 * Rethrows the passed error if it is a CancelledError
 */
function throwIfCancelledError(err) {
    if (isCancelledError(err)) {
        throw err;
    }
}
exports.throwIfCancelledError = throwIfCancelledError;
//# sourceMappingURL=cancellation.js.map