/// <reference types="lodash" />
import { Span } from 'opentracing';
import * as ts from 'typescript';
import { Disposable } from 'vscode-languageserver';
import { CancellationToken } from './cancellation';
import { FileSystemUpdater } from './fs';
import { Logger } from './logging';
import { InMemoryFileSystem } from './memfs';
/**
 * ProjectManager translates VFS files to one or many projects denoted by [tj]config.json.
 * It uses either local or remote file system to fetch directory tree and files from and then
 * makes one or more LanguageService objects. By default all LanguageService objects contain no files,
 * they are added on demand - current file for hover or definition, project's files for references and
 * all files from all projects for workspace symbols.
 */
export declare class ProjectManager implements Disposable {
    protected logger: Logger;
    /**
     * Cancellations to do when the object is disposed
     */
    private cancellationSources;
    /**
     * Root path (as passed to `initialize` request)
     */
    private rootPath;
    /**
     * Workspace subtree (folder) -> JS/TS configuration mapping.
     * Configuration settings for a source file A are located in the closest parent folder of A.
     * Map keys are relative (to workspace root) paths
     */
    private configs;
    /**
     * When on, indicates that client is responsible to provide file content (VFS),
     * otherwise we are working with a local file system
     */
    private strict;
    /**
     * Local side of file content provider which keeps cache of fetched files
     */
    private localFs;
    /**
     * File system updater that takes care of updating the in-memory file system
     */
    private updater;
    /**
     * Relative file path -> version map. Every time file content is about to change or changed (didChange/didOpen/...), we are incrementing it's version
     * signalling that file is changed and file's user must invalidate cached and requery file content
     */
    private versions;
    /**
     * Enables module resolution tracing by TS compiler
     */
    private traceModuleResolution;
    /**
     * Flag indicating that we fetched module struture (tsconfig.json, jsconfig.json, package.json files) from the remote file system.
     * Without having this information we won't be able to split workspace to sub-projects
     */
    ensuredModuleStructure?: Promise<void>;
    /**
     * Tracks if source file denoted by the given URI is fetched from remote file system and available locally.
     * For hover or definition we only need a single file (and maybe its transitive includes/references as reported by TS compiler).
     * This map prevents fetching of file content from remote filesystem twice
     */
    ensuredFilesForHoverAndDefinition: Map<string, Promise<void>>;
    /**
     * For references/symbols we need all the source files making workspace so this flag tracks if we already did it
     */
    private ensuredAllFiles?;
    /**
     * @param rootPath root path as passed to `initialize`
     * @param inMemoryFileSystem File system that keeps structure and contents in memory
     * @param strict indicates if we are working in strict mode (VFS) or with a local file system
     * @param traceModuleResolution allows to enable module resolution tracing (done by TS compiler)
     */
    constructor(rootPath: string, inMemoryFileSystem: InMemoryFileSystem, updater: FileSystemUpdater, strict: boolean, traceModuleResolution?: boolean, logger?: Logger);
    /**
     * Disposes the object and cancels any asynchronous operations that are still active
     */
    dispose(): void;
    /**
     * @return root path (as passed to `initialize`)
     */
    getRemoteRoot(): string;
    /**
     * @return local side of file content provider which keeps cached copies of fethed files
     */
    getFs(): InMemoryFileSystem;
    /**
     * @param filePath file path (both absolute or relative file paths are accepted)
     * @return true if there is a fetched file with a given path
     */
    hasFile(filePath: string): boolean;
    /**
     * @return all sub-projects we have identified for a given workspace.
     * Sub-project is mainly a folder which contains tsconfig.json, jsconfig.json, package.json,
     * or a root folder which serves as a fallback
     */
    getConfigurations(): ProjectConfiguration[];
    /**
     * ensureModuleStructure ensures that the module structure of the
     * project exists in localFs. TypeScript/JavaScript module
     * structure is determined by [jt]sconfig.json, filesystem layout,
     * global*.d.ts files. For performance reasons, we only read in
     * the contents of some files and store "var dummy_0ff1bd;" as the
     * contents of all other files.
     */
    ensureModuleStructure(): Promise<void>;
    /**
     * refreshFileTree refreshes the local in-memory filesytem's (this.localFs) files under the
     * specified path (root) with the contents of the remote filesystem (this.remoteFs). It will
     * also reset the ProjectConfigurations that are affected by the refreshed files.
     *
     * If moduleStructureOnly is true, then only files related to module structure (package.json,
     * tsconfig.json, etc.) will be refreshed.
     *
     * This method is public because a ProjectManager instance assumes there are no changes made to
     * the remote filesystem structure after initialization. If such changes are made, it is
     * necessary to call this method to alert the ProjectManager instance of the change.
     *
     * @param rootPath root path
     * @param moduleStructureOnly indicates if we need to fetch only configuration files such as tsconfig.json,
     * jsconfig.json or package.json (otherwise we want to fetch them plus source files)
     */
    refreshFileTree(rootPath: string, moduleStructureOnly: boolean): Promise<void>;
    /**
     * Ensures that all the files needed to produce hover and definitions for a given
     * source file URI were fetched from the remote file system. Set of the needed files includes:
     * - file itself
     * - file's includes and dependencies (transitive) reported by TS compiler up to depth 30
     * There is no need to fetch/parse/compile all the workspace files to produce hover of a symbol in the file F because
     * definition of this symbol must be in one of files references by F or its dependencies
     *
     * @param uri target file URI
     */
    ensureFilesForHoverAndDefinition(uri: string, childOf?: Span): Promise<void>;
    /**
     * Ensures all files needed for a workspace/symbol request are available in memory.
     * This includes all js/ts files, tsconfig files and package.json files.
     * It excludes files in node_modules.
     * Invalidates project configurations after execution
     */
    ensureFilesForWorkspaceSymbol: (() => Promise<void>) & _.MemoizedFunction;
    /**
     * Ensures all files were fetched from the remote file system.
     * Invalidates project configurations after execution
     */
    ensureAllFiles(): Promise<void>;
    /**
     * Ensures that we have all the files needed to retrieve all the references to a symbol in the given file.
     * Pretty much it's the same set of files needed to produce workspace symbols unless file is located in `node_modules`
     * in which case we need to fetch the whole tree
     *
     * @param uri target file URI
     */
    ensureFilesForReferences(uri: string): Promise<void>;
    /**
     * Recursively collects file(s) dependencies up to given level.
     * Dependencies are extracted by TS compiler from import and reference statements
     *
     * Dependencies include:
     * - all the configuration files
     * - files referenced by the given file
     * - files included by the given file
     *
     * @param filePaths files to process (both absolute and relative paths are accepted)
     * @param maxDepth stop collecting when reached given recursion level
     * @param seen tracks visited files to avoid cycles
     * @param childOf OpenTracing parent span for tracing
     */
    private ensureTransitiveFileDependencies(filePaths, maxDepth, seen?, childOf?);
    /**
     * @param filePath source file path relative to project root
     * @return project configuration for a given source file. Climbs directory tree up to workspace root if needed
     */
    getConfiguration(filePath: string): ProjectConfiguration;
    /**
     * Called when file was opened by client. Current implementation
     * does not differenciates open and change events
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didOpen(filePath: string, text: string): void;
    /**
     * Called when file was closed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     */
    didClose(filePath: string): void;
    /**
     * Called when file was changed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didChange(filePath: string, text: string): void;
    /**
     * Called when file was saved by client
     * @param filePath path to a file relative to project root
     */
    didSave(filePath: string): void;
    /**
     * ensureFiles ensures the following files have been fetched to
     * localFs. The files parameter is expected to contain paths in
     * the remote FS. ensureFiles only syncs unfetched file content
     * from remoteFs to localFs. It does not update project
     * state. Callers that want to do so after file contents have been
     * fetched should call this.createConfigurations().
     *
     * If one file fetch failed, the error will be caught and logged.
     *
     * @param files File paths
     */
    ensureFiles(files: Iterable<string>, token?: CancellationToken): Promise<void>;
    /**
     * Detects projects and creates projects denoted by tsconfig.json and jsconfig.json fiels.
     * Previously detected projects are NOT discarded.
     * If there is no root configuration, adds it to catch all orphan files
     */
    createConfigurations(): void;
}
/**
 * Implementaton of LanguageServiceHost that works with in-memory file system.
 * It takes file content from local cache and provides it to TS compiler on demand
 *
 * @implements ts.LanguageServiceHost
 */
export declare class InMemoryLanguageServiceHost implements ts.LanguageServiceHost {
    private logger;
    complete: boolean;
    /**
     * Root path
     */
    private rootPath;
    /**
     * Compiler options to use when parsing/analyzing source files.
     * We are extracting them from tsconfig.json or jsconfig.json
     */
    private options;
    /**
     * Local file cache where we looking for file content
     */
    private fs;
    /**
     * List of files that project consist of (based on tsconfig includes/excludes and wildcards).
     * Each item is a relative file path
     */
    expectedFilePaths: string[];
    /**
     * Current list of files that were implicitly added to project
     * (every time when we need to extract data from a file that we haven't touched yet).
     * Each item is a relative file path
     */
    private filePaths;
    /**
     * Current project version. When something significant is changed, incrementing it to signal TS compiler that
     * files should be updated and cached data should be invalidated
     */
    private projectVersion;
    /**
     * Tracks individual files versions to invalidate TS compiler data when single file is changed
     */
    private versions;
    constructor(rootPath: string, options: ts.CompilerOptions, fs: InMemoryFileSystem, expectedFiles: string[], versions: Map<string, number>, logger?: Logger);
    /**
     * TypeScript uses this method (when present) to compare project's version
     * with the last known one to decide if internal data should be synchronized
     */
    getProjectVersion(): string;
    /**
     * Incrementing current project version, telling TS compiler to invalidate internal data
     */
    incProjectVersion(): void;
    getCompilationSettings(): ts.CompilerOptions;
    getScriptFileNames(): string[];
    /**
     * Adds a file and increments project version, used in conjunction with getProjectVersion()
     * which may be called by TypeScript to check if internal data is up to date
     *
     * @param filePath relative file path
     */
    addFile(filePath: string): void;
    /**
     * @param fileName relative or absolute file path
     */
    getScriptVersion(fileName: string): string;
    /**
     * @param fileName relative or absolute file path
     */
    getScriptSnapshot(fileName: string): ts.IScriptSnapshot | undefined;
    getCurrentDirectory(): string;
    getDefaultLibFileName(options: ts.CompilerOptions): string;
    trace(message: string): void;
    log(message: string): void;
    error(message: string): void;
}
/**
 * ProjectConfiguration instances track the compiler configuration (as
 * defined by {tj}sconfig.json if it exists) and state for a single
 * TypeScript project. It represents the world of the view as
 * presented to the compiler.
 *
 * For efficiency, a ProjectConfiguration instance may hide some files
 * from the compiler, preventing them from being parsed and
 * type-checked. Depending on the use, the caller should call one of
 * the ensure* methods to ensure that the appropriate files have been
 * made available to the compiler before calling any other methods on
 * the ProjectConfiguration or its public members. By default, no
 * files are parsed.
 */
export declare class ProjectConfiguration {
    private logger;
    private service?;
    private program?;
    /**
     * Object TS service will use to fetch content of source files
     */
    private host?;
    /**
     * Local file cache
     */
    private fs;
    /**
     * Relative path to configuration file (tsconfig.json/jsconfig.json)
     */
    private configFilePath;
    /**
     * Configuration JSON object. May be used when there is no real configuration file to parse and use
     */
    private configContent;
    /**
     * Relative source file path (relative) -> version associations
     */
    private versions;
    /**
     * Enables module resolution tracing (done by TS service)
     */
    private traceModuleResolution;
    /**
     * Root file path, relative to workspace hierarchy root
     */
    private rootFilePath;
    /**
     * @param fs file system to use
     * @param rootFilePath root file path, relative to workspace hierarchy root
     * @param configFilePath configuration file path (relative to workspace root)
     * @param configContent optional configuration content to use instead of reading configuration file)
     */
    constructor(fs: InMemoryFileSystem, rootFilePath: string, versions: Map<string, number>, configFilePath: string, configContent?: any, traceModuleResolution?: boolean, logger?: Logger);
    /**
     * @return module resolution host to use by TS service
     */
    moduleResolutionHost(): ts.ModuleResolutionHost;
    /**
     * reset resets a ProjectConfiguration to its state immediately
     * after construction. It should be called whenever the underlying
     * local filesystem (fs) has changed, and so the
     * ProjectConfiguration can no longer assume its state reflects
     * that of the underlying files.
     */
    reset(): void;
    /**
     * @return package name (project name) of a given project
     */
    getPackageName(): string | null;
    /**
     * @return language service object
     */
    getService(): ts.LanguageService;
    /**
     * Note that it does not perform any parsing or typechecking
     * @return program object (cached result of parsing and typechecking done by TS service)
     */
    getProgram(): ts.Program;
    /**
     * @return language service host that TS service uses to read the data
     */
    getHost(): InMemoryLanguageServiceHost;
    /**
     * Tells TS service to recompile program (if needed) based on current list of files and compilation options.
     * TS service relies on information provided by language servide host to see if there were any changes in
     * the whole project or in some files
     */
    syncProgram(): void;
    private initialized?;
    /**
     * Initializes (sub)project by parsing configuration and making proper internal objects
     */
    private init();
    /**
     * Ensures we are ready to process files from a given sub-project
     */
    ensureConfigFile(): Promise<void>;
    private ensuredBasicFiles?;
    /**
     * Ensures we fetched basic files (global TS files, dependencies, declarations)
     */
    ensureBasicFiles(): Promise<void>;
    private ensuredAllFiles?;
    /**
     * Ensures we fetched all project's source file (as were defined in tsconfig.json)
     */
    ensureAllFiles(): Promise<void>;
}
