"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const iterare_1 = require("iterare");
const lodash_1 = require("lodash");
const opentracing_1 = require("opentracing");
const os = require("os");
const path_ = require("path");
const ts = require("typescript");
const cancellation_1 = require("./cancellation");
const logging_1 = require("./logging");
const memfs_1 = require("./memfs");
const util = require("./util");
/**
 * ProjectManager translates VFS files to one or many projects denoted by [tj]config.json.
 * It uses either local or remote file system to fetch directory tree and files from and then
 * makes one or more LanguageService objects. By default all LanguageService objects contain no files,
 * they are added on demand - current file for hover or definition, project's files for references and
 * all files from all projects for workspace symbols.
 */
class ProjectManager {
    /**
     * @param rootPath root path as passed to `initialize`
     * @param inMemoryFileSystem File system that keeps structure and contents in memory
     * @param strict indicates if we are working in strict mode (VFS) or with a local file system
     * @param traceModuleResolution allows to enable module resolution tracing (done by TS compiler)
     */
    constructor(rootPath, inMemoryFileSystem, updater, strict, traceModuleResolution, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        /**
         * Cancellations to do when the object is disposed
         */
        this.cancellationSources = new Set();
        /**
         * Tracks if source file denoted by the given URI is fetched from remote file system and available locally.
         * For hover or definition we only need a single file (and maybe its transitive includes/references as reported by TS compiler).
         * This map prevents fetching of file content from remote filesystem twice
         */
        this.ensuredFilesForHoverAndDefinition = new Map();
        /**
         * Ensures all files needed for a workspace/symbol request are available in memory.
         * This includes all js/ts files, tsconfig files and package.json files.
         * It excludes files in node_modules.
         * Invalidates project configurations after execution
         */
        this.ensureFilesForWorkspaceSymbol = lodash_1.memoize(() => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.updater.ensureStructure();
                const filesToEnsure = [];
                for (const uri of this.localFs.uris()) {
                    const file = util.uri2path(uri);
                    if (util.toUnixPath(file).indexOf('/node_modules/') === -1
                        && (util.isJSTSFile(file) || util.isConfigFile(file) || util.isPackageJsonFile(file))) {
                        filesToEnsure.push(file);
                    }
                }
                yield this.ensureFiles(filesToEnsure);
                yield this.createConfigurations();
            }
            catch (e) {
                this.ensureFilesForWorkspaceSymbol.cache = new WeakMap();
                throw e;
            }
        }));
        this.rootPath = util.toUnixPath(rootPath);
        this.configs = new Map();
        this.updater = updater;
        this.localFs = inMemoryFileSystem;
        this.versions = new Map();
        this.strict = strict;
        this.traceModuleResolution = traceModuleResolution || false;
    }
    /**
     * Disposes the object and cancels any asynchronous operations that are still active
     */
    dispose() {
        for (const source of this.cancellationSources) {
            source.cancel();
        }
    }
    /**
     * @return root path (as passed to `initialize`)
     */
    getRemoteRoot() {
        return this.rootPath;
    }
    /**
     * @return local side of file content provider which keeps cached copies of fethed files
     */
    getFs() {
        return this.localFs;
    }
    /**
     * @param filePath file path (both absolute or relative file paths are accepted)
     * @return true if there is a fetched file with a given path
     */
    hasFile(filePath) {
        return this.localFs.fileExists(filePath);
    }
    /**
     * @return all sub-projects we have identified for a given workspace.
     * Sub-project is mainly a folder which contains tsconfig.json, jsconfig.json, package.json,
     * or a root folder which serves as a fallback
     */
    getConfigurations() {
        const ret = [];
        this.configs.forEach((v, k) => {
            ret.push(v);
        });
        return ret;
    }
    /**
     * ensureModuleStructure ensures that the module structure of the
     * project exists in localFs. TypeScript/JavaScript module
     * structure is determined by [jt]sconfig.json, filesystem layout,
     * global*.d.ts files. For performance reasons, we only read in
     * the contents of some files and store "var dummy_0ff1bd;" as the
     * contents of all other files.
     */
    ensureModuleStructure() {
        if (!this.ensuredModuleStructure) {
            this.ensuredModuleStructure = this.refreshFileTree(this.rootPath, true).then(() => {
                this.createConfigurations();
            });
            this.ensuredModuleStructure.catch(err => {
                this.logger.error('Failed to fetch module structure: ', err);
                this.ensuredModuleStructure = undefined;
            });
        }
        return this.ensuredModuleStructure;
    }
    /**
     * refreshFileTree refreshes the local in-memory filesytem's (this.localFs) files under the
     * specified path (root) with the contents of the remote filesystem (this.remoteFs). It will
     * also reset the ProjectConfigurations that are affected by the refreshed files.
     *
     * If moduleStructureOnly is true, then only files related to module structure (package.json,
     * tsconfig.json, etc.) will be refreshed.
     *
     * This method is public because a ProjectManager instance assumes there are no changes made to
     * the remote filesystem structure after initialization. If such changes are made, it is
     * necessary to call this method to alert the ProjectManager instance of the change.
     *
     * @param rootPath root path
     * @param moduleStructureOnly indicates if we need to fetch only configuration files such as tsconfig.json,
     * jsconfig.json or package.json (otherwise we want to fetch them plus source files)
     */
    refreshFileTree(rootPath, moduleStructureOnly) {
        return __awaiter(this, void 0, void 0, function* () {
            rootPath = util.normalizeDir(rootPath);
            const filesToFetch = [];
            yield this.updater.fetchStructure();
            for (const uri of this.localFs.uris()) {
                const file = util.uri2path(uri);
                const rel = path_.posix.relative(this.rootPath, util.toUnixPath(file));
                if (!moduleStructureOnly || util.isGlobalTSFile(rel) || util.isConfigFile(rel) || util.isPackageJsonFile(rel)) {
                    filesToFetch.push(file);
                }
                else if (!this.localFs.fileExists(rel)) {
                    this.localFs.add(uri, localFSPlaceholder);
                }
            }
            yield this.ensureFiles(filesToFetch);
            // require re-fetching of dependency files (but not for
            // workspace/symbol and textDocument/references, because those
            // should not be affected by new external modules)
            this.ensuredFilesForHoverAndDefinition.clear();
            // require re-parsing of projects whose file set may have been affected
            for (let [dir, config] of this.configs) {
                dir = util.normalizeDir(dir);
                if (dir.startsWith(rootPath + '/') || rootPath.startsWith(dir + '/') || rootPath === dir) {
                    config.reset();
                }
            }
        });
    }
    /**
     * Ensures that all the files needed to produce hover and definitions for a given
     * source file URI were fetched from the remote file system. Set of the needed files includes:
     * - file itself
     * - file's includes and dependencies (transitive) reported by TS compiler up to depth 30
     * There is no need to fetch/parse/compile all the workspace files to produce hover of a symbol in the file F because
     * definition of this symbol must be in one of files references by F or its dependencies
     *
     * @param uri target file URI
     */
    ensureFilesForHoverAndDefinition(uri, childOf = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const span = childOf.tracer().startSpan('Ensure files for hover and definition', { childOf });
            span.addTags({ uri });
            try {
                const existing = this.ensuredFilesForHoverAndDefinition.get(uri);
                if (existing) {
                    return existing;
                }
                const promise = (() => __awaiter(this, void 0, void 0, function* () {
                    try {
                        yield this.ensureModuleStructure();
                        // Include dependencies up to depth 30
                        yield this.ensureTransitiveFileDependencies([util.uri2path(uri)], 30, undefined, span);
                    }
                    catch (err) {
                        this.ensuredFilesForHoverAndDefinition.delete(uri);
                        throw err;
                    }
                }))();
                this.ensuredFilesForHoverAndDefinition.set(uri, promise);
                yield promise;
            }
            catch (err) {
                span.setTag('error', true);
                span.log({ 'event': 'error', 'error.object': err });
                throw err;
            }
            finally {
                span.finish();
            }
        });
    }
    /**
     * Ensures all files were fetched from the remote file system.
     * Invalidates project configurations after execution
     */
    ensureAllFiles() {
        if (this.ensuredAllFiles) {
            return this.ensuredAllFiles;
        }
        const promise = this.updater.ensureStructure()
            .then(() => this.ensureFiles(iterare_1.default(this.localFs.uris())
            .map(uri => util.uri2path(uri))
            .filter(file => util.isJSTSFile(file))))
            .then(() => this.createConfigurations());
        this.ensuredAllFiles = promise;
        promise.catch(err => {
            this.logger.error('Failed to fetch files for references:', err);
            this.ensuredAllFiles = undefined;
        });
        return promise;
    }
    /**
     * Ensures that we have all the files needed to retrieve all the references to a symbol in the given file.
     * Pretty much it's the same set of files needed to produce workspace symbols unless file is located in `node_modules`
     * in which case we need to fetch the whole tree
     *
     * @param uri target file URI
     */
    ensureFilesForReferences(uri) {
        const fileName = util.uri2path(uri);
        if (util.toUnixPath(fileName).indexOf(`${path_.posix.sep}node_modules${path_.posix.sep}`) !== -1) {
            return this.ensureFilesForWorkspaceSymbol();
        }
        return this.ensureAllFiles();
    }
    /**
     * Recursively collects file(s) dependencies up to given level.
     * Dependencies are extracted by TS compiler from import and reference statements
     *
     * Dependencies include:
     * - all the configuration files
     * - files referenced by the given file
     * - files included by the given file
     *
     * @param filePaths files to process (both absolute and relative paths are accepted)
     * @param maxDepth stop collecting when reached given recursion level
     * @param seen tracks visited files to avoid cycles
     * @param childOf OpenTracing parent span for tracing
     */
    ensureTransitiveFileDependencies(filePaths, maxDepth, seen = new Set(), childOf = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const span = childOf.tracer().startSpan('Ensure file imports', { childOf });
            span.setTag('filePaths', filePaths.join(', '));
            try {
                filePaths = filePaths.filter(f => !seen.has(f));
                if (filePaths.length === 0) {
                    return Promise.resolve();
                }
                filePaths.forEach(f => seen.add(f));
                const absFilePaths = filePaths.map(f => util.toUnixPath(util.resolve(this.rootPath, f)));
                yield this.ensureFiles(absFilePaths);
                if (maxDepth > 0) {
                    const importPaths = new Set();
                    yield Promise.all(filePaths.map((filePath) => __awaiter(this, void 0, void 0, function* () {
                        const config = this.getConfiguration(filePath);
                        yield config.ensureBasicFiles();
                        const contents = this.getFs().readFile(filePath);
                        const info = ts.preProcessFile(contents, true, true);
                        const compilerOpt = config.getHost().getCompilationSettings();
                        for (const imp of info.importedFiles) {
                            const resolved = ts.resolveModuleName(util.toUnixPath(imp.fileName), filePath, compilerOpt, config.moduleResolutionHost());
                            if (!resolved || !resolved.resolvedModule) {
                                // This means we didn't find a file defining
                                // the module. It could still exist as an
                                // ambient module, which is why we fetch
                                // global*.d.ts files.
                                continue;
                            }
                            importPaths.add(resolved.resolvedModule.resolvedFileName);
                        }
                        const resolver = !this.strict && os.platform() === 'win32' ? path_ : path_.posix;
                        for (const ref of info.referencedFiles) {
                            // Resolving triple slash references relative to current file
                            // instead of using module resolution host because it behaves
                            // differently in "nodejs" mode
                            const refFilePath = util.toUnixPath(path_.relative(this.rootPath, resolver.resolve(this.rootPath, resolver.dirname(filePath), util.toUnixPath(ref.fileName))));
                            importPaths.add(refFilePath);
                        }
                    })));
                    yield this.ensureTransitiveFileDependencies(Array.from(importPaths), maxDepth - 1, seen, span);
                }
            }
            catch (err) {
                span.setTag('error', true);
                span.log({ 'event': 'error', 'error.object': err });
                throw err;
            }
            finally {
                span.finish();
            }
        });
    }
    /**
     * @param filePath source file path relative to project root
     * @return project configuration for a given source file. Climbs directory tree up to workspace root if needed
     */
    getConfiguration(filePath) {
        let dir = filePath;
        let config;
        while (dir && dir !== this.rootPath) {
            config = this.configs.get(dir);
            if (config) {
                return config;
            }
            dir = path_.posix.dirname(dir);
            if (dir === '.') {
                dir = '';
            }
        }
        config = this.configs.get('');
        if (config) {
            return config;
        }
        throw new Error(`TypeScript config file for ${filePath} not found`);
    }
    /**
     * Called when file was opened by client. Current implementation
     * does not differenciates open and change events
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didOpen(filePath, text) {
        this.didChange(filePath, text);
    }
    /**
     * Called when file was closed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     */
    didClose(filePath) {
        this.localFs.didClose(filePath);
        let version = this.versions.get(filePath) || 0;
        this.versions.set(filePath, ++version);
        const config = this.getConfiguration(filePath);
        config.ensureConfigFile().then(() => {
            config.getHost().incProjectVersion();
            config.syncProgram();
        });
    }
    /**
     * Called when file was changed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didChange(filePath, text) {
        this.localFs.didChange(filePath, text);
        let version = this.versions.get(filePath) || 0;
        this.versions.set(filePath, ++version);
        const config = this.getConfiguration(filePath);
        config.ensureConfigFile().then(() => {
            config.getHost().incProjectVersion();
            config.syncProgram();
        });
    }
    /**
     * Called when file was saved by client
     * @param filePath path to a file relative to project root
     */
    didSave(filePath) {
        this.localFs.didSave(filePath);
    }
    /**
     * ensureFiles ensures the following files have been fetched to
     * localFs. The files parameter is expected to contain paths in
     * the remote FS. ensureFiles only syncs unfetched file content
     * from remoteFs to localFs. It does not update project
     * state. Callers that want to do so after file contents have been
     * fetched should call this.createConfigurations().
     *
     * If one file fetch failed, the error will be caught and logged.
     *
     * @param files File paths
     */
    ensureFiles(files, token = cancellation_1.CancellationToken.None) {
        return __awaiter(this, void 0, void 0, function* () {
            const source = new cancellation_1.CancellationTokenSource();
            token.onCancellationRequested(() => source.cancel());
            this.cancellationSources.add(source);
            token = source.token;
            try {
                yield Promise.all(iterare_1.default(files).map((path) => __awaiter(this, void 0, void 0, function* () {
                    cancellation_1.throwIfRequested(token);
                    try {
                        yield this.updater.ensure(util.path2uri('', path));
                    }
                    catch (err) {
                        // if cancellation was requested, break out of the loop
                        cancellation_1.throwIfCancelledError(err);
                        cancellation_1.throwIfRequested(token);
                        // else log error and continue
                        this.logger.error(`Ensuring file ${path} failed`, err);
                    }
                })));
            }
            finally {
                this.cancellationSources.delete(source);
            }
        });
    }
    /**
     * Detects projects and creates projects denoted by tsconfig.json and jsconfig.json fiels.
     * Previously detected projects are NOT discarded.
     * If there is no root configuration, adds it to catch all orphan files
     */
    createConfigurations() {
        const rootdirs = new Set();
        for (const uri of this.localFs.uris()) {
            const relativeFilePath = path_.posix.relative(this.rootPath, util.uri2path(uri));
            if (!/(^|\/)[tj]sconfig\.json$/.test(relativeFilePath)) {
                continue;
            }
            if (/(^|\/)node_modules\//.test(relativeFilePath)) {
                continue;
            }
            let dir = path_.posix.dirname(relativeFilePath);
            if (dir === '.') {
                dir = '';
            }
            if (!this.configs.has(dir)) {
                this.configs.set(dir, new ProjectConfiguration(this.localFs, path_.posix.join('/', dir), this.versions, relativeFilePath, undefined, this.traceModuleResolution, this.logger));
            }
            rootdirs.add(dir);
        }
        if (!rootdirs.has('') && !this.configs.has('')) {
            // collecting all the files in workspace by making fake configuration object
            this.configs.set('', new ProjectConfiguration(this.localFs, '/', this.versions, '', {
                compilerOptions: {
                    module: ts.ModuleKind.CommonJS,
                    allowNonTsExtensions: false,
                    allowJs: true
                }
            }, this.traceModuleResolution, this.logger));
        }
    }
}
exports.ProjectManager = ProjectManager;
/**
 * Implementaton of LanguageServiceHost that works with in-memory file system.
 * It takes file content from local cache and provides it to TS compiler on demand
 *
 * @implements ts.LanguageServiceHost
 */
class InMemoryLanguageServiceHost {
    constructor(rootPath, options, fs, expectedFiles, versions, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        this.rootPath = rootPath;
        this.options = options;
        this.fs = fs;
        this.expectedFilePaths = expectedFiles;
        this.versions = versions;
        this.projectVersion = 1;
        this.filePaths = [];
    }
    /**
     * TypeScript uses this method (when present) to compare project's version
     * with the last known one to decide if internal data should be synchronized
     */
    getProjectVersion() {
        return '' + this.projectVersion;
    }
    /**
     * Incrementing current project version, telling TS compiler to invalidate internal data
     */
    incProjectVersion() {
        this.projectVersion++;
    }
    getCompilationSettings() {
        return this.options;
    }
    getScriptFileNames() {
        return this.filePaths;
    }
    /**
     * Adds a file and increments project version, used in conjunction with getProjectVersion()
     * which may be called by TypeScript to check if internal data is up to date
     *
     * @param filePath relative file path
     */
    addFile(filePath) {
        this.filePaths.push(filePath);
        this.incProjectVersion();
    }
    /**
     * @param fileName relative or absolute file path
     */
    getScriptVersion(fileName) {
        if (path_.posix.isAbsolute(fileName) || path_.isAbsolute(fileName)) {
            fileName = path_.posix.relative(this.rootPath, util.toUnixPath(fileName));
        }
        let version = this.versions.get(fileName);
        if (!version) {
            version = 1;
            this.versions.set(fileName, version);
        }
        return '' + version;
    }
    /**
     * @param fileName relative or absolute file path
     */
    getScriptSnapshot(fileName) {
        let exists = this.fs.fileExists(fileName);
        if (!exists) {
            fileName = path_.posix.join(this.rootPath, fileName);
            exists = this.fs.fileExists(fileName);
        }
        if (!exists) {
            return undefined;
        }
        return ts.ScriptSnapshot.fromString(this.fs.readFile(fileName));
    }
    getCurrentDirectory() {
        return this.rootPath;
    }
    getDefaultLibFileName(options) {
        return util.toUnixPath(ts.getDefaultLibFilePath(options));
    }
    trace(message) {
        // empty
    }
    log(message) {
        // empty
    }
    error(message) {
        this.logger.error(message);
    }
}
exports.InMemoryLanguageServiceHost = InMemoryLanguageServiceHost;
const localFSPlaceholder = 'var dummy_0ff1bd;';
/**
 * ProjectConfiguration instances track the compiler configuration (as
 * defined by {tj}sconfig.json if it exists) and state for a single
 * TypeScript project. It represents the world of the view as
 * presented to the compiler.
 *
 * For efficiency, a ProjectConfiguration instance may hide some files
 * from the compiler, preventing them from being parsed and
 * type-checked. Depending on the use, the caller should call one of
 * the ensure* methods to ensure that the appropriate files have been
 * made available to the compiler before calling any other methods on
 * the ProjectConfiguration or its public members. By default, no
 * files are parsed.
 */
class ProjectConfiguration {
    /**
     * @param fs file system to use
     * @param rootFilePath root file path, relative to workspace hierarchy root
     * @param configFilePath configuration file path (relative to workspace root)
     * @param configContent optional configuration content to use instead of reading configuration file)
     */
    constructor(fs, rootFilePath, versions, configFilePath, configContent, traceModuleResolution, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        this.fs = fs;
        this.configFilePath = configFilePath;
        this.configContent = configContent;
        this.versions = versions;
        this.traceModuleResolution = traceModuleResolution || false;
        this.rootFilePath = rootFilePath;
    }
    /**
     * @return module resolution host to use by TS service
     */
    moduleResolutionHost() {
        return this.fs;
    }
    /**
     * reset resets a ProjectConfiguration to its state immediately
     * after construction. It should be called whenever the underlying
     * local filesystem (fs) has changed, and so the
     * ProjectConfiguration can no longer assume its state reflects
     * that of the underlying files.
     */
    reset() {
        this.initialized = undefined;
        this.ensuredBasicFiles = undefined;
        this.ensuredAllFiles = undefined;
        this.service = undefined;
        this.program = undefined;
        this.host = undefined;
    }
    /**
     * @return package name (project name) of a given project
     */
    getPackageName() {
        // package.json may be located at the upper level as well
        let currentDir = this.rootFilePath;
        while (true) {
            const pkgJsonFile = path_.posix.join(currentDir, 'package.json');
            if (this.fs.fileExists(pkgJsonFile)) {
                return JSON.parse(this.fs.readFile(pkgJsonFile)).name;
            }
            const parentDir = path_.dirname(currentDir);
            if (parentDir === '.' || parentDir === '/' || parentDir === currentDir) {
                return null;
            }
            currentDir = parentDir;
        }
    }
    /**
     * @return language service object
     */
    getService() {
        if (!this.service) {
            throw new Error('project is uninitialized');
        }
        return this.service;
    }
    /**
     * Note that it does not perform any parsing or typechecking
     * @return program object (cached result of parsing and typechecking done by TS service)
     */
    getProgram() {
        if (!this.program) {
            throw new Error('project is uninitialized');
        }
        return this.program;
    }
    /**
     * @return language service host that TS service uses to read the data
     */
    getHost() {
        if (!this.host) {
            throw new Error('project is uninitialized');
        }
        return this.host;
    }
    /**
     * Tells TS service to recompile program (if needed) based on current list of files and compilation options.
     * TS service relies on information provided by language servide host to see if there were any changes in
     * the whole project or in some files
     */
    syncProgram() {
        this.program = this.getService().getProgram();
    }
    /**
     * Initializes (sub)project by parsing configuration and making proper internal objects
     */
    init() {
        if (this.initialized) {
            return this.initialized;
        }
        this.initialized = new Promise((resolve, reject) => {
            let configObject;
            if (!this.configContent) {
                const jsonConfig = ts.parseConfigFileTextToJson(this.configFilePath, this.fs.readFile(this.configFilePath));
                if (jsonConfig.error) {
                    this.logger.error('Cannot parse ' + this.configFilePath + ': ' + jsonConfig.error.messageText);
                    return reject(new Error('Cannot parse ' + this.configFilePath + ': ' + jsonConfig.error.messageText));
                }
                configObject = jsonConfig.config;
            }
            else {
                configObject = this.configContent;
            }
            let dir = path_.posix.dirname(this.configFilePath);
            if (dir === '.') {
                dir = '';
            }
            const base = dir || this.fs.path;
            const configParseResult = ts.parseJsonConfigFileContent(configObject, this.fs, base);
            const expFiles = configParseResult.fileNames;
            // Add globals that might exist in dependencies
            const nodeModulesDir = path_.posix.join(base, 'node_modules');
            const err = memfs_1.walkInMemoryFs(this.fs, nodeModulesDir, (path, isdir) => {
                if (!isdir && util.isGlobalTSFile(path)) {
                    expFiles.push(path);
                }
            });
            if (err) {
                return reject(err);
            }
            const options = configParseResult.options;
            if (/(^|\/)jsconfig\.json$/.test(this.configFilePath)) {
                options.allowJs = true;
            }
            if (this.traceModuleResolution) {
                options.traceResolution = true;
            }
            this.host = new InMemoryLanguageServiceHost(this.fs.path, options, this.fs, expFiles, this.versions, this.logger);
            this.service = ts.createLanguageService(this.host, ts.createDocumentRegistry());
            this.program = this.service.getProgram();
            return resolve();
        });
        return this.initialized;
    }
    /**
     * Ensures we are ready to process files from a given sub-project
     */
    ensureConfigFile() {
        return this.init();
    }
    /**
     * Ensures we fetched basic files (global TS files, dependencies, declarations)
     */
    ensureBasicFiles() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.ensuredBasicFiles) {
                return this.ensuredBasicFiles;
            }
            this.ensuredBasicFiles = this.init().then(() => {
                let changed = false;
                for (const fileName of (this.getHost().expectedFilePaths || [])) {
                    if (util.isGlobalTSFile(fileName) || (!util.isDependencyFile(fileName) && util.isDeclarationFile(fileName))) {
                        const sourceFile = this.getProgram().getSourceFile(fileName);
                        if (!sourceFile) {
                            this.getHost().addFile(fileName);
                            changed = true;
                        }
                    }
                }
                if (changed) {
                    // requery program object to synchonize LanguageService's data
                    this.program = this.getService().getProgram();
                }
            });
            return this.ensuredBasicFiles;
        });
    }
    /**
     * Ensures we fetched all project's source file (as were defined in tsconfig.json)
     */
    ensureAllFiles() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.ensuredAllFiles) {
                return this.ensuredAllFiles;
            }
            this.ensuredAllFiles = this.init().then(() => {
                if (this.getHost().complete) {
                    return;
                }
                let changed = false;
                for (const fileName of (this.getHost().expectedFilePaths || [])) {
                    const sourceFile = this.getProgram().getSourceFile(fileName);
                    if (!sourceFile) {
                        this.getHost().addFile(fileName);
                        changed = true;
                    }
                }
                if (changed) {
                    // requery program object to synchonize LanguageService's data
                    this.program = this.getService().getProgram();
                }
                this.getHost().complete = true;
            });
            return this.ensuredAllFiles;
        });
    }
}
exports.ProjectConfiguration = ProjectConfiguration;
//# sourceMappingURL=project-manager.js.map